﻿//        Copyright (c) 2009-2011, Semyon A. Chertkov (semyonc@gmail.com)
//        All rights reserved.
//
//        This program is free software: you can redistribute it and/or modify
//        it under the terms of the GNU General Public License as published by
//        the Free Software Foundation, either version 3 of the License, or
//        any later version.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Globalization;

using System.Linq;
using System.Xml;
using System.Xml.XPath;
using System.Xml.Schema;
using System.Xml.Linq;

using DataEngine.CoreServices;
using DataEngine.XQuery.Parser;
using DataEngine.XQuery.Util;

namespace DataEngine.XQuery
{
    public static class XPathFactory
    {
        public static XPathItem Clone(this XPathItem item)
        {
            XPathNavigator nav = item as XPathNavigator;
            if (nav != null)
                return nav.Clone();
            else
                return ((XQueryItemBase)item).Clone();
        }

        public static XQueryNodeIterator Query(this XPathNavigator node, string xquery, IXmlNamespaceResolver nsmgr)
        {
            XPathContext context = new XPathContext(node.NameTable);
            XQueryCommand command = new XQueryCommand(context);
            if (nsmgr != null)
                context.CopyNamespaces(nsmgr);
            command.ContextItem = node;
            command.CommandText = xquery;
            return command.Execute();
        }

        public static XQueryNodeIterator Query(this XPathNavigator node, string xquery)
        {
            return Query(node, xquery, null);
        }

        internal static object CloneObject(this Object src)
        {
            XPathItem item = src as XPathItem;
            if (item != null)
                return item.Clone();
            XQueryNodeIterator iter = src as XQueryNodeIterator;
            if (iter != null)
                return iter.Clone();
            return src;
        }

        private class NodeList : XmlNodeList
        {
            private List<XmlNode> _list;
            private XQueryNodeIterator _iter;
            private bool _done;
            private DOMConverter _converter;

            public NodeList(XQueryNodeIterator iter, XmlDocument doc)
            {
                _list = new List<XmlNode>();
                _iter = iter;
                _converter = new DOMConverter(doc);
            }

            public override int Count
            {
                get
                {
                    if (!_done)
                        Item(Int32.MaxValue);
                    return _list.Count;
                }
            }

            private XmlNode GetNode(XPathItem item)
            {
                if (item.IsNode)
                    return _converter.ToXmlNode((XPathNavigator)item);
                return null;
            }

            public override IEnumerator GetEnumerator()
            {
                return new Enumerator(this);
            }

            public override XmlNode Item(int index)
            {
                if (_list.Count <= index && !_done)
                {
                    int count = _list.Count;
                    while (!_done && (count <= index))
                    {
                        if (_iter.MoveNext())
                        {
                            XmlNode node = GetNode(_iter.Current);
                            if (node != null)
                            {
                                _list.Add(node);
                                count++;
                            }
                        }
                        else
                            _done = true;
                    }
                }
                if (index >= 0 && _list.Count > index)
                    return _list[index];
                return null;
            }

            private class Enumerator : IEnumerator, IEnumerator<XmlNode>
            {
                private NodeList _owner;
                private XmlNode _curr;
                private int _pos = -1;

                public Enumerator(NodeList owner)
                {
                    _owner = owner;
                }

                #region IEnumerator Members

                object IEnumerator.Current
                {
                    get
                    {
                        if (_pos == -1)
                            throw new InvalidOperationException();
                        return _curr;
                    }
                }

                bool IEnumerator.MoveNext()
                {
                    XmlNode node = _owner[_pos + 1];
                    if (node != null)
                    {
                        _curr = node;
                        _pos++;
                        return true;
                    }
                    return false;
                }

                void IEnumerator.Reset()
                {
                    _pos = -1;
                }

                #endregion

                #region IEnumerator<XmlNode> Members

                XmlNode IEnumerator<XmlNode>.Current
                {
                    get
                    {
                        if (_pos == -1)
                            throw new InvalidOperationException();
                        return _curr;
                    }
                }

                #endregion

                #region IDisposable Members

                void IDisposable.Dispose()
                {
                    return;
                }

                #endregion
            }

        }

        public static XmlNodeList QueryNodes(this XmlNode node, string xquery)
        {
            return QueryNodes(node, xquery, null);
        }

        public static XmlNodeList QueryNodes(this XmlNode node, string xquery, IXmlNamespaceResolver nsmgr)
        {
            XPathNavigator nav = node.CreateNavigator();
            XPathContext context = new XPathContext(nav.NameTable);
            XQueryCommand command = new XQueryCommand(context);
            if (nsmgr != null)
                context.CopyNamespaces(nsmgr);
            command.ContextItem = nav;
            command.CommandText = xquery;
            XmlDocument owner = node as XmlDocument;
            if (owner == null)
                owner = node.OwnerDocument;
            return new NodeList(command.Execute(), owner);
        }

        public static XmlNode QuerySingleNode(this XmlNode node, string xquery)
        {
            return QuerySingleNode(node, xquery, null);
        }

        public static XmlNode QuerySingleNode(this XmlNode node, string xquery, IXmlNamespaceResolver nsResolver)
        {
            XmlNodeList nodes = QueryNodes(node, xquery, nsResolver);
            return nodes[0];
        }

        public static IEnumerable<T> XQuerySelect<T>(this XNode node, string xquery)
            where T : XObject
        {
            return XQuerySelect<T>(node, xquery, null);
        }

        public static IEnumerable<T> XQuerySelect<T>(this XNode node, string xquery, IXmlNamespaceResolver nsResolver)
            where T : XObject
        {
            return XQueryCommand.Select<T>(xquery, nsResolver, node);
        }

        public static IEnumerable<Object> XQuerySelect(this XNode node, string xquery, IXmlNamespaceResolver nsResolver)
        {
            return XQueryCommand.Select(xquery, nsResolver, node);
        }

        public static IEnumerable<Object> XQuerySelect(this XNode node, string xquery)
        {
            return XQuerySelect(node, xquery, null);
        }

        public static T XQuerySelectOne<T>(this XNode node, string xquery, IXmlNamespaceResolver nsResolver)
            where T : XObject
        {
            return XQueryCommand.SelectOne<T>(xquery, nsResolver, node);
        }

        public static T XQuerySelectOne<T>(this XNode node, string xquery)
            where T : XObject
        {
            return XQuerySelectOne<T>(node, xquery, null);
        }

        public static Object XQuerySelectOne(this XNode node, string xquery, IXmlNamespaceResolver nsResolver)
        {
            return XQueryCommand.SelectOne(xquery, nsResolver, node);
        }

        public static Object XQuerySelectOne(this XNode node, string xquery)
        {
            return XQuerySelectOne(node, xquery, null);
        }

        public static IEnumerable<XElement> XQuerySelectElements(this XNode node, string xquery, IXmlNamespaceResolver nsResolver)
        {
            return XQuerySelect<XElement>(node, xquery, nsResolver);
        }

        public static IEnumerable<XElement> XQuerySelectElements(this XNode node, string xquery)
        {
            return XQuerySelectElements(node, xquery, null);
        }

        public static XElement XQuerySelectElement(this XNode node, string xquery, IXmlNamespaceResolver nsResolver)
        {
            return XQuerySelectOne<XElement>(node, xquery, nsResolver);
        }

        public static XElement XQuerySelectElement(this XNode node, string xquery)
        {
            return XQuerySelectElement(node, xquery, null);
        }

        public static Type GetNavigatorValueType(XPathNavigator nav, Type valueType)
        {
            IXmlSchemaInfo schemaInfo = nav.SchemaInfo;
            if (schemaInfo == null || schemaInfo.SchemaType == null)
                return typeof(UntypedAtomic);
            switch (schemaInfo.SchemaType.TypeCode)
            {
                case XmlTypeCode.UntypedAtomic:
                    return typeof(UntypedAtomic);
                case XmlTypeCode.Integer:
                case XmlTypeCode.PositiveInteger:
                case XmlTypeCode.NegativeInteger:
                case XmlTypeCode.NonPositiveInteger:
                    return typeof(Integer);
                case XmlTypeCode.Date:
                    return typeof(DateValue);
                case XmlTypeCode.DateTime:
                    return typeof(DateTimeValue);
                case XmlTypeCode.Time:
                    return typeof(TimeValue);
                case XmlTypeCode.Duration:
                    return typeof(DurationValue);
                case XmlTypeCode.DayTimeDuration:
                    return typeof(DayTimeDurationValue);
                case XmlTypeCode.YearMonthDuration:
                    return typeof(YearMonthDurationValue);
                case XmlTypeCode.GDay:
                    return typeof(GDayValue);
                case XmlTypeCode.GMonth:
                    return typeof(GMonthValue);
                case XmlTypeCode.GMonthDay:
                    return typeof(GMonthDayValue);
                case XmlTypeCode.GYear:
                    return typeof(GYearValue);
                case XmlTypeCode.GYearMonth:
                    return typeof(GYearMonthValue);
                case XmlTypeCode.QName:
                    return typeof(QNameValue);
                case XmlTypeCode.Notation:
                    return typeof(NotationValue);
                case XmlTypeCode.AnyUri:
                    return typeof(AnyUriValue);
                case XmlTypeCode.HexBinary:
                    return typeof(HexBinaryValue);
                case XmlTypeCode.Base64Binary:
                    return typeof(Base64BinaryValue);
                case XmlTypeCode.Idref:
                    if (schemaInfo.SchemaType == XQuerySequenceType.XmlSchema.IDREFS)
                        return typeof(IDREFSValue);
                    goto default;
                case XmlTypeCode.NmToken:
                    if (schemaInfo.SchemaType == XQuerySequenceType.XmlSchema.NMTOKENS)
                        return typeof(NMTOKENSValue);
                    goto default;
                case XmlTypeCode.Entity:
                    if (schemaInfo.SchemaType == XQuerySequenceType.XmlSchema.ENTITIES)
                        return typeof(ENTITIESValue);
                    goto default;
                default:
                    return valueType;
            }
        }

        internal static XPathItem ChangeType(this XPathItem item, XQuerySequenceType destType, XQueryContext context)
        {
            if (destType.IsNode)
            {
                if (!destType.Match(item, context))
                    throw new XQueryException(Properties.Resources.XPTY0004,
                        new XQuerySequenceType(item.XmlType.TypeCode), destType);
                return item.Clone();
            }
            else
            {
                if (destType.SchemaType == item.XmlType)
                    return item.Clone();
                else if (destType.TypeCode == XmlTypeCode.Item &&
                    (destType.Cardinality == XmlTypeCardinality.One || destType.Cardinality == XmlTypeCardinality.ZeroOrOne))
                    return item.Clone();
                else
                {
                    XmlSchemaSimpleType simpleType = destType.SchemaType as XmlSchemaSimpleType;
                    if (simpleType == null)
                        throw new InvalidOperationException();
                    if (simpleType == XQuerySequenceType.XmlSchema.AnySimpleType)
                        throw new XQueryException(Properties.Resources.XPST0051, "xs:anySimpleType");
                    return new XQueryItem(XQueryConvert.ChangeType(item.XmlType, item.TypedValue,
                        destType, context.nameTable, context.NamespaceManager), destType.SchemaType);
                }
            }
        }

        internal static IEnumerable<XPathItem> RootIterator(XQueryNodeIterator iter)
        {
            foreach (XPathItem item in iter)
            {
                XPathNavigator nav = item as XPathNavigator;
                if (nav != null)
                {
                    XPathNavigator curr = nav.Clone();
                    curr.MoveToRoot();
                    yield return curr;
                }
            }
        }

        internal static IEnumerable<XPathItem> AttributeIterator(XQueryNodeIterator iter)
        {
            foreach (XPathItem item in iter)
            {
                XPathNavigator nav = item as XPathNavigator;
                if (nav != null)
                {
                    XPathNavigator curr = nav.Clone();
                    if (curr.MoveToFirstAttribute())
                        do
                        {
                            yield return curr;
                        } while (curr.MoveToNextAttribute());
                }
            }
        }

        internal static IEnumerable<XPathItem> UnionIterator1(XQueryNodeIterator iter1, XQueryNodeIterator iter2)
        {
            SortedDictionary<XPathItem, XPathItem> set =
                new SortedDictionary<XPathItem, XPathItem>(new XPathComparer());
            foreach (XPathItem item in iter1)
                if (!set.ContainsKey(item))
                    set.Add(item.Clone(), null);
            foreach (XPathItem item in iter2)
                if (!set.ContainsKey(item))
                    set.Add(item.Clone(), null);
            foreach (KeyValuePair<XPathItem, XPathItem> kvp in set)
                yield return kvp.Key;
        }

        internal static IEnumerable<XPathItem> UnionIterator2(XQueryNodeIterator iter1, XQueryNodeIterator iter2)
        {
            HashSet<XPathItem> hs = new HashSet<XPathItem>(new XPathNavigatorEqualityComparer());
            foreach (XPathItem item in iter1)
                if (!hs.Contains(item))
                {
                    hs.Add(item.Clone());
                    yield return item;
                }
            foreach (XPathItem item in iter2)
                if (!hs.Contains(item))
                {
                    hs.Add(item.Clone());
                    yield return item;
                }
        }

        internal static IEnumerable<XPathItem> IntersectExceptIterator1(bool except, XQueryNodeIterator iter1, XQueryNodeIterator iter2)
        {
            SortedDictionary<XPathItem, XPathItem> set =
                new SortedDictionary<XPathItem, XPathItem>(new XPathComparer());
            HashSet<XPathItem> hs = new HashSet<XPathItem>(new XPathNavigatorEqualityComparer());
            foreach (XPathItem item in iter1)
                if (!set.ContainsKey(item))
                    set.Add(item.Clone(), null);
            foreach (XPathItem item in iter2)
                if (!hs.Contains(item))
                    hs.Add(item.Clone());
            foreach (KeyValuePair<XPathItem, XPathItem> kvp in set)
                if (except)
                {
                    if (!hs.Contains(kvp.Key))
                        yield return kvp.Key;
                }
                else
                {
                    if (hs.Contains(kvp.Key))
                        yield return kvp.Key;
                }
        }

        internal static IEnumerable<XPathItem> IntersectExceptIterator2(bool except, XQueryNodeIterator iter1, XQueryNodeIterator iter2)
        {
            HashSet<XPathItem> hs = new HashSet<XPathItem>(new XPathNavigatorEqualityComparer());
            foreach (XPathItem item in iter1)
                if (!hs.Contains(item))
                    hs.Add(item.Clone());
            if (except)
                hs.ExceptWith(iter2);
            else
                hs.IntersectWith(iter2);
            foreach (XPathItem item in hs)
                yield return item;
        }

        internal static IEnumerable<XPathItem> ConvertIterator(XQueryNodeIterator iter, XQuerySequenceType destType, XQueryContext context)
        {
            int num = 0;
            XQuerySequenceType itemType = new XQuerySequenceType(destType);
            itemType.Cardinality = XmlTypeCardinality.One;
            foreach (XPathItem item in iter)
            {
                if (num == 1)
                {
                    if (destType.Cardinality == XmlTypeCardinality.ZeroOrOne ||
                        destType.Cardinality == XmlTypeCardinality.One)
                        throw new XQueryException(Properties.Resources.XPTY0004, "item()+", destType);
                }
                yield return item.ChangeType(itemType, context);
                num++;
            }
            if (num == 0)
            {
                if (destType.Cardinality == XmlTypeCardinality.One ||
                    destType.Cardinality == XmlTypeCardinality.OneOrMore)
                    throw new XQueryException(Properties.Resources.XPTY0004, "item()?", destType);
            }
        }

        internal static IEnumerable<XPathItem> ValueIterator(XQueryNodeIterator iter, XQuerySequenceType destType,
            XQueryContext context)
        {
            int num = 0;
            foreach (XPathItem item in iter)
            {
                if (num == 1)
                {
                    if (destType.Cardinality == XmlTypeCardinality.ZeroOrOne ||
                        destType.Cardinality == XmlTypeCardinality.One)
                        throw new XQueryException(Properties.Resources.XPTY0004, "item()+", destType);
                }
                if (destType.IsNode)
                {
                    if (!destType.Match(item, context))
                        throw new XQueryException(Properties.Resources.XPTY0004,
                            new XQuerySequenceType(item.XmlType, XmlTypeCardinality.OneOrMore, null), destType);
                    yield return item;
                }
                else
                    yield return new XQueryItem(XQueryConvert.ValueAs(item.TypedValue, destType,
                        context.nameTable, context.NamespaceManager));
                num++;
            }
            if (num == 0)
            {
                if (destType.Cardinality == XmlTypeCardinality.One ||
                    destType.Cardinality == XmlTypeCardinality.OneOrMore)
                    throw new XQueryException(Properties.Resources.XPTY0004, "item()?", destType);
            }
        }

        internal static IEnumerable<XPathItem> TreatIterator(XQueryNodeIterator iter, XQuerySequenceType destType, XQueryContext context)
        {
            int num = 0;
            foreach (XPathItem item in iter)
            {
                if (num == 1)
                {
                    if (destType.Cardinality == XmlTypeCardinality.ZeroOrOne ||
                        destType.Cardinality == XmlTypeCardinality.One)
                        throw new XQueryException(Properties.Resources.XPTY0004, "item()+", destType);
                }
                if (destType.IsNode)
                {
                    if (!destType.Match(item, context))
                        throw new XQueryException(Properties.Resources.XPTY0004,
                            new XQuerySequenceType(item.XmlType, XmlTypeCardinality.OneOrMore, null), destType);
                    yield return item;
                }
                else
                    yield return new XQueryItem(XQueryConvert.TreatValueAs(item.TypedValue, destType));
                num++;
            }
            if (num == 0)
            {
                if (destType.Cardinality == XmlTypeCardinality.One ||
                    destType.Cardinality == XmlTypeCardinality.OneOrMore)
                    throw new XQueryException(Properties.Resources.XPTY0004, "item()?", destType);
            }
        }

        internal static IEnumerable<XPathItem> ValidateIterator(XQueryNodeIterator iter, XmlSchemaSet schemaSet, bool lax)
        {
            int n = 0;
            foreach (XPathItem item in iter)
            {
                if (!item.IsNode)
                    throw new XQueryException(Properties.Resources.XPTY0004,
                        new XQuerySequenceType(item.TypedValue.GetType(), XmlTypeCardinality.One), "node()*");
                XPathNavigator nav = (XPathNavigator)item.Clone();
                try
                {
                    nav.CheckValidity(schemaSet, null);
                }
                catch (XmlSchemaValidationException ex)
                {
                    throw new XQueryException(ex.Message, ex);
                }
                catch (InvalidOperationException ex)
                {
                    throw new XQueryException(ex.Message, ex);
                }
                yield return nav;
                n++;
            }
            if (n == 0)
                throw new XQueryException(Properties.Resources.XQTY0030);
        }

        internal static IEnumerable<XPathItem> CodepointIterator(string text)
        {
            for (int k = 0; k < text.Length; k++)
                yield return new XQueryItem(Convert.ToInt32(text[k]));
        }


        // Prevent (attribute attr {})/.., (text {})/.., etc expressions
        internal static IEnumerable<XPathItem> DynAttributeIterator(XPathNavigator nav)
        {
            XQueryNavigator curr = (XQueryNavigator)nav.Clone();
            if (curr.MoveToFirstAttribute())
                do
                {
                    yield return curr;
                } while (curr.MoveToNextAttribute());
        }
    }
}
